﻿/**
Copyright(c)2008 nutsu.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/**
* Quadratic Bezier Utility
* @author nutsu
* @version 0.2
*/

package com.nutsu.shape{
	
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import com.nutsu.geom.LineF;
	import com.nutsu.geom.Line;
	
	public class QBezierUtil{
		
		//t値誤差
		private static var ERR_T:Number = 1e-5;
		
		
		/**
		* 2次ベジェ曲線と点との最短距離
		* @param Bezier
		* @param Point
		* @return 距離
		*/
		public static function distanceFromPoint( bezier:QBezier, p:Point ):Number{
			var t:Number = nearestFromPoint( bezier, p );
			var dv:Point = ( bezier.f(t) ).subtract( p );
			return dv.length;
		}
		
		/**
		* 2次ベジェ曲線と点との最短距離 t値
		* @param Bezier
		* @param Point
		* @return t値
		*/
		public static function nearestFromPoint( bezier:QBezier, p:Point ):Number{
			
			//bezierを(-p)でオフセットした座標
			var p0:Point  = bezier.point0;
			var p1:Point  = bezier.point1;
			var pc:Point  = bezier.control;
			
			var x0:Number = p0.x - p.x;
			var y0:Number = p0.y - p.y;
			var x1:Number = p1.x - p.x;
			var y1:Number = p1.y - p.y;
			var cx:Number = pc.x - p.x;
			var cy:Number = pc.y - p.y;
			
			//計算パラメータ
			var xx:Number = x0+x1;
			var yy:Number = y0+y1;
			var cc:Number = cx*cx + cy*cy;
			
			//曲線と点の距離平方の微分計算定数
			// fd(t) = A*t^3 + B*t^2 + C*t + D
			var A:Number  = 4*( xx*xx + yy*yy + 4*cc - 4*cx*xx - 4*cy*yy );
			var B:Number  = -12*( x0*xx + y0*yy + 2*cc + (-3*x0-x1)*cx + (-3*y0-y1)*cy );
			var C:Number  = 4*( (3*x0+x1-6*cx)*x0 + (3*y0+y1-6*cy)*y0 + 2*cc );
			var D:Number  = -4*(y0*y0-cy*y0+x0*x0-cx*x0);
			
			var TA:Number = 27*A*A*D*D + (4*B*B*B-18*A*B*C)*D + 4*A*C*C*C - B*B*C*C;
			
			
			if( TA>0 ){
				//実数解
				TA = Math.sqrt(TA)/( 6*Math.sqrt(3)*A*A ) - (27*A*A*D - 9*A*B*C + 2*B*B*B )/(54*A*A*A);
				if( TA<0 ){
					TA = -Math.pow(-TA,1/3);
				}else{
					TA = Math.pow(TA,1/3);
				}
				return Math.max( 0.0, Math.min( 1.0, (TA - (3*A*C-B*B)/(9*A*A*TA) - B/(3*A)) ) );
				
			}else if( TA<0 ){
				//虚数解あり、数値計算で
				
				//距離平方の微分 ※2次関数の平方の和なので、Aは正→山・谷の3次式
				var diff:Function = function(t:Number):Number{ return (A*t*t*t + B*t*t + C*t + D); };
				
				//2回微分「3*A*t*t + 2*B*t + C」の解
				var tt0:Number = (-B - Math.sqrt(B*B-3*A*C))/(3*A);
				var tt1:Number = (-B + Math.sqrt(B*B-3*A*C))/(3*A)
				
				//最適解
				var t0:Number = 0;
				var t1:Number = 1.0;
				
				var ot0:Number;
				var ot1:Number;
				var dt:Number;
				
				if( tt0>0 && diff(0)<0 ){
					//数値計算
					ot0 = 0.0;
					ot1 = tt0;
					t0  = (ot0 + ot1)/2;
					while( Math.abs( dt=diff(t0) )>ERR_T ){
						if( dt>0 ){
							ot1 = t0;
							t0  = (ot0 + ot1)/2;
						}else{
							ot0 = t0;
							t0  = (ot0 + ot1)/2;
						}
					}
				}
				if( tt1<1.0 && diff(1.0)>0 ){
					//数値計算
					ot0 = tt1;
					ot1 = 1.0;
					t1  = (ot0 + ot1)/2;
					while( Math.abs( dt=diff(t1) )>ERR_T ){
						if( dt<0 ){
							ot0 = t1;
							t1  = (ot0 + ot1)/2;
						}else{
							ot1 = t1;
							t1  = (ot0 + ot1)/2;
						}
					}
				}
				
				var dv0:Point = new Point( t0*t0*(x1+x0-2*cx)+t0*(2*cx-2*x0)+x0, t0*t0*(y1+y0-2*cy)+t0*(2*cy-2*y0)+y0 );
				var dv1:Point = new Point( t1*t1*(x1+x0-2*cx)+t1*(2*cx-2*x0)+x0, t1*t1*(y1+y0-2*cy)+t1*(2*cy-2*y0)+y0 );
				
				//等距離の時はどうするか？
				return ( dv0.length<dv1.length ) ? t0 : t1;
				
			}else{
				//コントロール点が、始点と終点の中点
				return -D/C;
				
			}
		}
		
		/**
		* ベジェ曲線と直線の交点のt値
		* 
		* 接している場合に誤差あり（接点が2点になるのはよいが、接していないことになる場合は難あり）
		* 
		* @param	LineF
		* @return	t値[]
		*/
		public static function intersectionTByLineF( bezier:QBezier, sl:LineF, include_tangent:Boolean=true ):Array{
			var x0:Number = bezier.x0;
			var y0:Number = bezier.y0;
			var x1:Number = bezier.x1;
			var y1:Number = bezier.y1;
			var cx:Number = bezier.cx;
			var cy:Number = bezier.cy;
			
			//ax+by+c=0の式にベジェ関数を代入 → mt^2+nt+l=0
			var a:Number = sl.a; var b:Number = sl.b;
			var m:Number = a*( x0 + x1 - 2*cx ) + b*(y0 + y1 - 2*cy);
			var n:Number = 2*( a*cx + b*cy - a*x0 - b*y0 );
			var l:Number = b*y0 + a*x0 + sl.c;
	
			//一次式になる場合
			if( m==0 ){
				var ta:Number = -l/n;
				if( ta>-1e-10 && ta<(1+1e-10) ){
					return [ Math.max(0,Math.min(ta,1)) ];
				}else{
					return [];
				}
			}
			
			//判別式
			var D:Number = n*n - 4*m*l;
			if( D>0 ){
				D = Math.sqrt(D);
				
				//2次式の解
				var t0:Number = 0.5*(-n-D)/m;
				var t1:Number = 0.5*(-n+D)/m;
				
				var result:Array = new Array();
				
				//解が0～1にあれば交点
				if( Math.abs(t1-t0)<1e-7 ){
					//接点とみなす
					if( t0>-1e-10 && t0<(1+1e-10) ){
						t0 = Math.max(0,Math.min(t0,1));
						if( t0<1e-10 ){
							result.push( 0 );
						}else if( t0>1-1e-10 ){
							result.push( 1 );
						}else if( include_tangent ){
							result.push( t0 );
						}
					}
				}else{
					if( t0>-1e-10 && t0<(1+1e-10) ){
						result.push( Math.max(0,Math.min(t0,1)) );
					}
					if( t1>-1e-10 && t1<(1+1e-10) ){
						result.push( Math.max(0,Math.min(t1,1)) );
					}
				}
				return result;
				
			}else if( D==0 && include_tangent ){
				var t2:Number = 0.5*-n/m;
				if( t2>-1e-10 && t2<(1+1e-10) ){
					return [ Math.max(0,Math.min(t2,1)) ];
				}else{
					return []
				}
			}else{
				//交点なし
				return [];
			}
		}
		
		/**
		* ベジェ曲線と線分の交点のt値
		* @param	Line
		* @return	t値[]
		*/
		public static function intersectionTByLine( bezier:QBezier, sl:Line, include_tangent:Boolean=true ):Array{
			var ts:Array = intersectionTByLineF( bezier, sl.f, include_tangent );
			var rs:Array = new Array();
			for( var i:int=0; i<ts.length; i++ ){
				var p:Point = bezier.f( ts[i] );
				if( ((p.x-sl.point0.x)*(p.x-sl.point1.x)+(p.y-sl.point0.y)*(p.y-sl.point1.y))<=0 ){
					rs.push( ts[i] );
				}
			}
			return rs;
		}
		
		/**
		* ベジェ曲線とRectの交点のt値(接点は含めない・端点が接点の場合含める)
		* @param	Rectangle
		* @return	t値[]
		*/
		public static function intersectionTByRect( bezier:QBezier, rt:Rectangle ):Array{
			var p0:Point = rt.topLeft;
			var p1:Point = new Point(rt.right,rt.top);
			var p2:Point = rt.bottomRight;
			var p3:Point = new Point(rt.left,rt.bottom);
			//cal intersections
			var sl0:Line = new Line( p0, p1 );
			var sl1:Line = new Line( p1, p2 );
			var sl2:Line = new Line( p2, p3 );
			var sl3:Line = new Line( p3, p0 );
			var ts:Array = intersectionTByLine(bezier,sl0,false).concat(intersectionTByLine(bezier,sl1,false)).concat(intersectionTByLine(bezier,sl2,false)).concat(intersectionTByLine(bezier,sl3,false));
			ts.sort(Array.NUMERIC);
			//頂点での交点重複を削除
			for( var i:int=1; i<ts.length; i++ ){
				if( Math.abs(ts[i]-ts[i-1])<1e-10 ){
					ts.splice( i,1 );
					i--;
				}
			}
			return ts;
		}
		
		/**
		 * 2次ベジェの交点
		 * @param	QBezier
		 * @param	QBezier
		 * @return	Point[]
		 */
		public static function intersection( b0:QBezier, b1:QBezier, th:Number=0.1 ):Array{
			var bs:Array = b1.split(0.5);
			var ps:Array = seekIntersection(b0,bs[0],th).concat(seekIntersection(b0,bs[1],th));
			return ps;
		}
		
		private static function seekIntersection( b0:QBezier, b1:QBezier, th:Number):Array{
			var res:Array = new Array();
			var ts:Array;
			
			//Clip By FatLine
			var r:Number = Math.atan2(b1.y1-b1.y0,b1.x1-b1.x0);
			var s:Number = Math.sin(-r);
			var c:Number = Math.cos(-r);
			//
			var br0:QBezier = rotateBezier( b0, s, c );
			var br1:QBezier = rotateBezier( b1, s, c );
			//
			var fatlinesize:Rectangle = br1.getRect();
			ts = br0.clipTValueByRect(fatlinesize);
			
			//check and next seek
			if( ts.length==2 && Math.min(fatlinesize.width,fatlinesize.height)<th ){
				res.push( b0.f( (ts[0]+ts[1])/2 ) );
			}else{
				for( var i:int=0; i<ts.length ;i+=2 ){
					res = res.concat( seekIntersection( b1, b0.clip(ts[i],ts[i+1]), th ) );
				}
			}
			return res;
		}
		
		/**
		 * 曲線の回転
		 */
		private static function rotateBezier( b:QBezier, sin:Number, cos:Number ):QBezier{
			var x0:Number = b.x0;
			var y0:Number = b.y0;
			var x1:Number = b.x1;
			var y1:Number = b.y1;
			var cx:Number = b.cx;
			var cy:Number = b.cy;
			return new QBezier( cos*x0-sin*y0, sin*x0+cos*y0, cos*x1-sin*y1, sin*x1+cos*y1, cos*cx-sin*cy, sin*cx+cos*cy );
		}
	}
	
	
}


